from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, Any, Union, Optional, Type, Sequence
from functools import wraps

from ..reqeust import Response, BaseRequest


__all__ = (
    "HTTPClient",
    "ResourceData",
    "ErrorData",
    "SemesterData",
    "APIResponseType",
    "IParser",
    "ILoginFetcher",
    "IGeneralFetcher",
    "ISemesterFetcher",
    "ParserPrecondition",
)


class HTTPClient:
    connector: Optional[Type[BaseRequest]] = None

    @classmethod
    def set_auto(cls):
        http: Type[BaseRequest]
        for http in BaseRequest.__subclasses__():
            cls.connector = http

    @classmethod
    def set(cls, connector: Type[BaseRequest]):
        cls.connector = connector


@dataclass
class ResourceData:
    data: Dict[str, Any]
    link: str
    meta: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ErrorData:
    error: Dict[str, Any]
    link: str
    meta: Dict[str, Any] = field(default_factory=dict)


@dataclass
class SemesterData:
    selected: str
    selectable: Sequence[str]


APIResponseType = Union[ResourceData, ErrorData]


class IParser(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def parse(cls, response: Response) -> APIResponseType:
        """ html 또는 raw_data 를 사용가능한 데이터로 변환

        ParsingError: 의도되지 않은 parse 실패
        """
        pass


class ILoginFetcher(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    async def fetch(
        cls,
        user_id: str,
        user_pw: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        pass


class IGeneralFetcher(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        pass


class ISemesterFetcher(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        semester: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        pass


class ParserPrecondition:
    def __init__(self, baseclass):
        self.baseclass = baseclass

    def __call__(self, func):
        @wraps(func)
        def wrapper(cls, response: Response):
            for subclass in self.baseclass.__subclasses__():
                error = subclass.is_blocking(response)
                if error:
                    return error
            return func(cls, response)

        return wrapper


HTTPClient.set_auto()



from .base import (
    HTTPClient,
    ResourceData,
    ErrorData,
)
from .intranet import Login as IntranetLogin
from .intranet import StudentPhoto as IntranetStudentPhoto
from .intranet import Chapel as IntranetChapel
from .intranet import Timetable as IntranetTimetable
from .intranet import Course as IntranetCourse
from .intranet import TotalAcceptanceStatus as IntranetTotalAcceptanceStatus
from .intranet import GraduationExam as IntranetGraduationExam
from .intranet import Profile as IntranetProfile
from .lms import Login as LmsLogin
from .lms import Profile as LmsProfile
from .lms import CourseList as LmsCourseList
from .lms import Attendance as LmsAttendance
from .kbu import MainNotice as KbuMainNotice
from .kbu import ScholarshipNotice as KbuScholarshipNotice
from .kbu import IllipNotice as KbuIllipNotice
from .mileage import Login as MileageLogin
from .mileage import Search as MileageSearch
from .mileage import Statement as MileageStatement
from ._mileage import SearchParamData as MileageSearchParamData
from ._mileage import StatementParamData as MileageStatementParamData
from .library import Login as LibraryLogin
from .library import CheckoutList as LibraryCheckoutList
from .library import BookDetail as LibraryBookDetail
from .library import BookPhoto as LibraryBookPhoto


__all__ = (
    "HTTPClient",
    "ResourceData",
    "ErrorData",
    "IntranetAPI",
    "LmsAPI",
    "KbuAPI",
    "MileageAPI",
    "MileageParam",
    "LibraryAPI",
)


class IntranetAPI:
    Login = IntranetLogin
    StudentPhoto = IntranetStudentPhoto
    Chapel = IntranetChapel
    Timetable = IntranetTimetable
    Course = IntranetCourse
    TotalAcceptanceStatus = IntranetTotalAcceptanceStatus
    GraduationExam = IntranetGraduationExam
    Profile = IntranetProfile


class LmsAPI:
    Login = LmsLogin
    Profile = LmsProfile
    CourseList = LmsCourseList
    Attendance = LmsAttendance


class KbuAPI:
    MainNotice = KbuMainNotice
    ScholarshipNotice = KbuScholarshipNotice
    IllipNotice = KbuIllipNotice


class MileageAPI:
    Login = MileageLogin
    Search = MileageSearch
    Statement = MileageStatement


class MileageParam:
    SearchParamData = MileageSearchParamData
    StatementParamData = MileageStatementParamData


class LibraryAPI:
    Login = LibraryLogin
    CheckoutList = LibraryCheckoutList
    BookDetail = LibraryBookDetail
    BookPhoto = LibraryBookPhoto



"""
##### 대출목록 기능 #####
CheckoutList → BookDetail → BookPhoto를 거쳐야만 데이터가 취합되는 구조이다.
CheckoutList를 통해
['No', '서지정보', '대출일자', '반납예정일', '대출상태', '연기신청', '상세페이지 URL']의 데이터가

BookDetail과 BookPhoto를 통해
['ISBN', '서지정보', '대출일자', '반납예정일', '대출상태', '연기신청', '도서이미지']의 데이터가 완성된다.
"""
from typing import Dict, Optional, List, Tuple
from base64 import b64encode
import re

from .base import (
    ILoginFetcher,
    IParser,
    HTTPClient,
    APIResponseType,
    ResourceData,
    ErrorData,
    ParserPrecondition,
)
from ..reqeust.base import Response
from ..api.intranet import IParserPrecondition
from .common import (
    httpdate_to_unixtime,
    parse_table,
    extract_alerts
)

__all__ = (
    "Login",
    "CheckoutList",
    "BookDetail",
    "BookPhoto",
)

DOMAIN_NAME: str = "https://lib.bible.ac.kr"

_ParserPrecondition = ParserPrecondition(IParserPrecondition)


class _SessionExpiredChecker(IParserPrecondition):
    @staticmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        if response.status == 302:
            return ErrorData(
                error={"title": "세션이 만료되어 로그인페이지로 리다이렉트 되었습니다."}, link=response.url
            )
        return None


class Login(ILoginFetcher, IParser):
    URL: str = DOMAIN_NAME + "/Account/LogOn"

    @classmethod
    async def fetch(
        cls,
        user_id: str,
        user_pw: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        form = {
            "l_id": b64encode(user_id.encode()).decode(),
            "l_pass": b64encode(user_pw.encode()).decode(),
        }
        return await HTTPClient.connector.post(
            cls.URL, headers=headers, body=form, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        """
        로그인 성공: status 302, location header 포함, 리다이렉트 메세지를 body에 포함
        로그인 실패: status 200, location header 미포함, alert 메세지를 body에 포함
            [상황 1] 잘못된 입력값
            [상황 2] 서비스 이용 불가한 졸업예정자
            [상황 3] 인코딩, 잘못된 경로 입력
        """
        soup = response.soup
        # 로그인 성공
        if response.status == 302:
            iat = httpdate_to_unixtime(response.headers["date"])
            return ResourceData(
                data={"cookies": response.cookies, "iat": iat}, link=response.url
            )
        # 로그인 실패: [상황 2] 일시 [상항 1] alert 보다 먼저 알림
        if "location" not in response.headers:
            alerts: List[str] = extract_alerts(response.soup)
            alerts.append(soup.select_one(".alert-warning").text.strip())
            alert = alerts[0] if alerts else ""
            return ErrorData(
                error={"title": alert, "alert_messages": alerts}, link=response.url
            )
        # 로그인 실패: [상황 3]
        m = re.search(r"ErrorCode=(\d+)", response.headers["location"])
        if m:
            error_code = m.group(1)
            return ErrorData(
                error={"title": "잘못된 경로입니다.", "code": error_code}, link=response.url
            )


class CheckoutList(IParser):
    URL: str = DOMAIN_NAME + "/MyLibrary"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            cls.URL, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List, List]:
        soup = response.soup
        thead = soup.select_one(".sponge-table-default thead")
        tbody = soup.select_one(".sponge-table-default tbody")
        head, body = parse_table(response, thead, tbody)

        del head[4]
        head[-1] = "상세페이지 URL"

        for tr, each in zip(tbody.select("tr"), body):
            del each[4]
            each[1] = tr.select_one("td a strong").text.strip()
            each[-1] = tr.select_one(".left a")["href"]
        return head, body

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        head, body = cls._parse_main_table(response)
        if not body:
            return ErrorData(error={"title": "대출한 내역이 없습니다."}, link=response.url)

        return ResourceData(data={"head": head, "body": body}, link=response.url)


class BookDetail:
    @classmethod
    async def fetch(
        cls,
        path,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            DOMAIN_NAME + path, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> List[str]:
        soup = response.soup
        isbn = soup.select("#detailtoprightnew .sponge-book-list-data")[1].text.strip()
        img_url = soup.select_one(".page-detail-title-image a img")["src"]

        if not re.match(r"https?://", img_url):
            img_url = None
        return [isbn, img_url]


class BookPhoto:
    @classmethod
    async def fetch(
        cls,
        photo_url,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            photo_url, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        if response.headers["content-type"][:5] == "image":
            return ResourceData(data={"raw_image": response.raw}, link=response.url)



from .base import (
    BaseRequest,
    Response,
    HTTPRequestMethod,
    IRequestPostCondition,
)

try:
    from .aiohttp_conn import Request
except ImportError:
    pass

try:
    from .requests_conn import Request
except ImportError:
    pass

__all__ = (
    "BaseRequest",
    "IRequestPostCondition",
    "Response",
    "HTTPRequestMethod",
)



import re
from typing import List, Mapping, Dict, Tuple
import urllib.parse
import datetime

from bs4 import BeautifulSoup
import bs4.element

from ..exceptions import ParsingError
from ..reqeust import Response


__all__ = (
    "httpdate_to_unixtime",
    "extract_alerts",
    "extract_hidden_tags",
    "urlencode",
    "remove_unexpected_char",
    "SemesterConverter",
    "parse_table",
)

_ALERT_PATTERN = re.compile(r"[^a-zA-Z0-9_]*?alert\s*\((.+?)\)")


def httpdate_to_unixtime(date: str) -> int:
    return int(
        (
            datetime.datetime.strptime(date.strip(), "%a, %d %b %Y %H:%M:%S GMT")
            - datetime.datetime(1970, 1, 1)
        ).total_seconds()
    )


def _replace_alert_message(message: str):
    message = message.strip()
    f, e = message[0], message[-1]
    if f == e and (f == "'" or f == '"'):
        return message.strip(f)


def extract_alerts(soup: BeautifulSoup) -> List[str]:
    script_elements = soup.find_all("script")

    result = []
    for each in script_elements:
        each = each.text if each.text else str(each.string)

        for message in _ALERT_PATTERN.findall(each):
            alert = _replace_alert_message(message)
            if alert:
                result.append(alert)
    return result


def extract_hidden_tags(soup: BeautifulSoup) -> Dict[str, str]:
    hidden_tags = soup.find_all("input", type="hidden")
    return {tag.get("name"): tag.get("value", "") for tag in hidden_tags}


def urlencode(queries: Mapping[str, str], encoding: str = "utf-8") -> str:
    return urllib.parse.urlencode(queries, encoding=encoding)


def remove_unexpected_char(text: str) -> str:
    return (
        text.strip()
        .replace("\t", "")
        .replace("\n", "")
        .replace("\r", "")
        .replace(u"\xa0", u"")
    )


class SemesterConverter:
    SEMESTER: Tuple[Tuple[str, str]] = (
        ("1", "10"),  # 1학기
        ("2", "20"),  # 2학기
        ("3", "11"),  # 여름 계절학기
        ("4", "21"),  # 겨울 계절학기
    )
    INTRANET_STYLE: int = 0
    LMS_STYLE: int = 1

    @classmethod
    def _convert_semester(cls, semester: str, style: int) -> str:
        for each in cls.SEMESTER:
            if semester == each[style]:
                return each[(style + 1) % 2]

    @classmethod
    def intranet_to_lms(cls, semester: str) -> Dict[str, str]:
        year = semester[:4]
        semester = cls._convert_semester(semester[4], cls.INTRANET_STYLE)

        return {"year": year, "semester": semester}

    @classmethod
    def lms_to_intranet(cls, year: str, semester: str) -> str:
        semester = cls._convert_semester(semester, cls.LMS_STYLE)

        return year + semester


def parse_table(
    response: Response, thead: bs4.element.Tag, tbody: bs4.element.Tag,
) -> Tuple[List[str], List[List[str]]]:
    if not thead:
        raise ParsingError("테이블 헤드가 존재하지 않습니다.", response)
    if not tbody:
        raise ParsingError("테이블 바디가 존재하지 않습니다.", response)

    head: List[str] = [th.get_text(strip=True) for th in thead.find_all("th")]
    body: List[List[str]] = [
        [td.get_text(strip=True) for td in tr.find_all("td")]
        for tr in tbody.find_all("tr")
    ]
    return head, body



from .__version__ import *
from .api import *
from .reqeust import *
from .exceptions import *

__all__ = (
    "__title__",
    "__description__",
    "__url__",
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "HTTPClient",
    "ResourceData",
    "ErrorData",
    "IntranetAPI",
    "LmsAPI",
    "KbuAPI",
    "MileageAPI",
    "MileageParam",
    "BaseRequest",
    "IRequestPostCondition",
    "Response",
    "HTTPRequestMethod",
    "RootError",
    "RequestError",
    "ResponseError",
    "RequestTimeoutError",
    "ClientError",
    "ServerError",
    "ParsingError",
    "LibraryAPI",
)



from dataclasses import dataclass
from typing import List, Tuple, Dict

__all__ = (
    "translate_mileage_req",
    "translate_statement_type",
    "SearchParamData",
    "StatementParamData",
)


_REQ_MEANINGS: Dict[str, str] = {
    "CST_NO": "회원번호",
    "CST_NM": "회원명",
    "CST_CARD_NO": "카드번호",
    "PROD_CD": "상품코드",
    "PROD_NM": "상품명",
    "dataSeq": "dataSeq",
    "SALE_QTY": "판매수량",
    "TOT_SALE_AMT": "총매출",
    "TOT_DC_AMT": "할인액",
    "DCM_SALE_AMT": "순매출",
    "SALE_DATE": "판매일자",
    "SALE_TIME": "판매시간",
    "SHOP_NM": "매장명",
    "CST_USE_POINT": "사용포인트",
    "CHG_DATE": "변경일자",
    "CHG_FG": "구분",
    "POINT": "포인트",
    "REMARK": "비고",
    "CST_CLS_NM": "회원등급",
    "BIRTH_DATE": "생일",
    "TEL_NO": "전화번호",
    "HP_NO": "휴대폰번호",
    "SMS_RECV_YN": "SMS수신여부",
    "HP_CHK": "HP_CHK",
    "SMS_REJECT_YN": "SMS_REJECT_YN",
    "USESHOP_ACC_POINT": "USESHOP_ACC_POINT",
    "USESHOP_USE_POINT": "USESHOP_USE_POINT",
    "USESHOP_ADJ_POINT": "USESHOP_ADJ_POINT",
    "USESHOP_ACC_SALE_CNT": "USESHOP_ACC_SALE_CNT",
    "USESHOP_ACC_SALE_AMT": "USESHOP_ACC_SALE_AMT",
    "USESHOP_F_SALE_DATE": "USESHOP_F_SALE_DATE",
    "USESHOP_L_SALE_DATE": "USESHOP_L_SALE_DATE",
    "ACC_POINT": "포인트적립",
    "USE_POINT": "포인트사용",
    "ADJ_POINT": "포인트조정",
    "AVL_POINT": "포인트가용",
    "ACC_SALE_CNT": "결제횟수",
    "ACC_SALE_AMT": "결제금액",
    "F_SALE_DATE": "최초방문일",
    "L_SALE_DATE": "최종방문일",
    "INS_DT": "가입일",
    "TERM_ACC_POINT": "TERM_ACC_POINT",
    "TERM_USE_POINT": "TERM_USE_POINT",
    "TERM_ADJ_POINT": "TERM_ADJ_POINT",
    "TERM_AVL_POINT": "TERM_AVL_POINT",
    "ADDR": "주소",
    "DM_RECV_YN": "DM수신여부",
    "CST_ID": "회원참조",
}

_STATEMENT_TYPES: Dict[str, str] = {
    "0": "신규",
    "1": "이관",
    "2": "조정",
    "3": "적립",
    "4": "적립취소",
    "5": "사용",
    "6": "사용취소",
}


def translate_mileage_req(request_information_str: str) -> List[str]:
    return [
        _REQ_MEANINGS.get(each, each) for each in request_information_str.split("|")
    ]


def translate_statement_type(statement_type: str) -> str:
    return _STATEMENT_TYPES.get(statement_type, statement_type)


@dataclass
class SearchParamData:
    birth_day: str = "01"
    birth_day2: str = "01"
    birth_month: str = "01"
    birth_month2: str = "01"
    cst_nos: str = ""
    date1_1: str = "2019-08-12"  # 과거 날짜여도 상관 없음
    date1_2: str = "2019-08-12"  # 과거 날짜여도 상관 없음
    date_period1: str = "366"
    list: str = ""
    mySheet1: str = ""
    page_no: str = "1"  # ✓
    page_size: str = "5000"  # ✓
    page_url: str = "/master/cust/cust020.jsp"
    r_ogn_cd: str = "HNON"
    r_ogn_fg: str = "C"
    row_cnt: str = ""
    CHG_SHOP_CD: str = ""
    CST_CARD_NO: str = ""  # ✓ 학번 검색
    CST_CARD_USE_FG: str = ""
    CST_CLS_CD: str = ""
    CST_ID: str = ""
    CST_NM: str = ""
    CST_NO: str = ""  # # ✓ 회원번호 검색
    DATE_FG: str = "A"
    DM_RECV_YN: str = ""
    EMAIL_ADDR: str = ""
    EMAIL_RECV_YN: str = ""
    EX_CST: str = ""
    EX_VISIT: str = ""
    HP_CARD1: str = ""
    HP_CARD2: str = ""
    HP_NO1: str = ""  # ✓ 010/011/016
    HP_NO2: str = ""  # ✓ aaaa/aaa
    HP_NO3: str = ""  # ✓ bbbb
    HP_SALE1: str = ""
    HP_SALE2: str = ""
    INS_SHOP_CD: str = ""
    MVIEW_YN: str = "N"
    ORDERYN: str = ""
    PNT_FG: str = "A"
    SALE_FG: str = "C"
    SEX_FG: str = "A"
    SHEETSEQ: str = "1"
    SMS_RECV_YN: str = ""
    S_CONTROLLER: str = "master.cust.cust020"
    S_CST_NM: str = ""
    S_FORWARD: str = ""
    S_METHOD: str = "search"
    S_SAVENAME: str = "CST_NO|CST_NM|CST_CLS_NM|CST_CARD_NO|ACC_POINT|USE_POINT|ADJ_POINT|AVL_POINT|ACC_SALE_CNT|ACC_SALE_AMT|INS_DT"
    S_TEL_NO: str = ""
    S_TREECOL: str = ""
    SaleYN: str = ""
    TEL_NO1: str = ""
    TEL_NO2: str = ""
    TEL_NO3: str = ""
    TERM_FG: str = "A"
    USE_YN: str = ""
    WEDDING_YN: str = "A"

    def get_req(self) -> str:
        return self.S_SAVENAME

    def set_req(self, req_info: str):
        self.S_SAVENAME = req_info
        return self

    def get_student_id(self) -> str:
        return self.CST_CARD_NO

    def set_student_id(self, student_id: str):
        self.CST_CARD_NO = student_id
        return self

    def get_customer_id(self) -> str:
        return self.CST_NO

    def set_customer_id(self, customer_id: str):
        self.CST_NO = customer_id
        return self

    def get_page_num(self) -> str:
        return self.page_no

    def set_page_num(self, num: str):
        self.page_no = num
        return self

    def get_page_size(self) -> str:
        return self.page_size

    def set_page_size(self, size: str):
        self.page_size = size
        return self

    def get_phone_number(self) -> Tuple[str, str, str]:
        return self.HP_NO1, self.HP_NO2, self.HP_NO3

    def set_phone_number(self, a: str, b: str, c: str):
        self.HP_NO1 = a
        self.HP_NO2 = b
        self.HP_NO3 = c
        return self


@dataclass
class StatementParamData:
    S_CONTROLLER: str = "master.cust.cust020_cst_info"
    S_METHOD: str = "search"
    S_SAVENAME: str = "CHG_DATE|SALE_DATE|CHG_FG|POINT|REMARK"
    SHEETSEQ: str = "2"
    CST_NO: str = ""  # ✓ 고객번호
    page_no: str = "1"
    page_no2: str = "1"
    page_size: str = "2000"
    page_size2: str = "2000"

    def get_req(self) -> str:
        return self.S_SAVENAME

    def set_req(self, req_info: str):
        self.S_SAVENAME = req_info
        return self

    def get_customer_id(self) -> str:
        return self.CST_NO

    def set_customer_id(self, customer_id: str):
        self.CST_NO = customer_id
        return self

    def get_page_num(self) -> str:
        return self.page_no

    def set_page_num(self, num: str):
        self.page_no = num
        return self

    def get_page_size(self) -> str:
        return self.page_size

    def set_page_size(self, size: str):
        self.page_size = size
        return self



from abc import ABCMeta, abstractmethod
from typing import Optional, Dict, List, Tuple
from collections import defaultdict
import re

from .base import (
    HTTPClient,
    IParser,
    APIResponseType,
    ILoginFetcher,
    IGeneralFetcher,
    ISemesterFetcher,
    ResourceData,
    ErrorData,
    ParserPrecondition,
    SemesterData,
)
from ..reqeust import Response
from ..exceptions import ParsingError
from .common import (
    httpdate_to_unixtime,
    extract_alerts,
    extract_hidden_tags,
    urlencode,
    parse_table,
)

__all__ = (
    "IParserPrecondition",
    "Login",
    "StudentPhoto",
    "Chapel",
    "Timetable",
    "Course",
    "TotalAcceptanceStatus",
    "GraduationExam",
)

DOMAIN_NAME: str = "https://kbuis.bible.ac.kr"  # with protocol
_SEMESTER_KEY: str = "ctl00$ContentPlaceHolder1$cbo_YearHg"


class IParserPrecondition(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        """ 진행할 수 없는 사전조건인 경우 ErrorData, 그렇지 않은 경우 None """
        pass


_ParserPrecondition = ParserPrecondition(IParserPrecondition)


class _SessionExpiredChecker(IParserPrecondition):
    @staticmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        alerts = extract_alerts(response.soup)
        for alert in alerts:
            if "세션" in alert or "수업평가" in alert:
                return ErrorData(
                    error={"title": alert, "alert_messages": alerts}, link=response.url
                )
        return None


def _extract_semester(response: Response) -> SemesterData:
    select_tag = response.soup.find("select", attrs={"name": _SEMESTER_KEY})
    if not select_tag:
        raise ParsingError("학기 셀렉트 태그를 찾을 수 없습니다.", response)
    options = select_tag.find_all("option", selected=True)
    if not options:
        raise ParsingError("학기 옵션 태그를 찾을 수 없습니다.", response)

    try:
        selectables: List[str] = [
            opt.attrs["value"] for opt in select_tag.find_all("option")
        ]
        selected: str = select_tag.find("option", selected=True).attrs["value"]
    except (KeyError, AttributeError):
        raise ParsingError("학기 옵션 태그를 정상적으로 선택할 수 없습니다.", response)
    return SemesterData(selected=selected, selectable=selectables)


async def _post_with_semester(
    url,
    cookies: Dict[str, str],
    semester: Optional[str] = None,
    *,
    headers: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> Response:
    """ 인트라넷에서 특정 학기의 정보 조회를 위한 메서드

    특정 학기 조회를 위해서는 POST 메서드로 정보를 전송해야하는데, 그 전에 hidden 태그를 함께 보내야함.
    1. GET 요청, 해당 페이지를 불러와서 form hidden-tag 의 (name,key) 쌍을 얻는다.
        - 여기서 얻는 정보는 학교에서 미리 지정해놓은터 학기, 일반적으로 최신 학기
    2. POST 요청, hidden-tag와 학기를 body에 담아 전송한다.
    """
    response = await HTTPClient.connector.get(
        url, cookies=cookies, headers=headers, timeout=timeout, **kwargs
    )
    if _SessionExpiredChecker.is_blocking(response):
        return response

    semester_info: SemesterData = _extract_semester(response)
    if (
        semester
        and semester != semester_info.selected
        and semester in semester_info.selectable
    ):
        body = extract_hidden_tags(response.soup)
        body[_SEMESTER_KEY] = semester
        body["ctl00$ContentPlaceHolder1$hidActionMode"] = "S"
        response = await HTTPClient.connector.post(
            url, body=body, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )
        semester_info: SemesterData = _extract_semester(response)
    response.etc["semester"] = semester_info
    return response


class Login(ILoginFetcher, IParser):
    URL: str = DOMAIN_NAME + "/ble_login3.aspx"

    @classmethod
    async def _get_extra_payload(
        cls
    ) -> Tuple[str, str]:
        response = await HTTPClient.connector.get(cls.URL)

        soup = response.soup
        view_state = soup.find("input", {"id": "__VIEWSTATE"}).get('value')
        event_validation = soup.find("input", {"id": "__EVENTVALIDATION"}).get('value')

        if not (view_state or event_validation):
            raise ParsingError("페이 로드에 파싱할 extra 값이 없습니다. ", response)

        return view_state, event_validation

    @classmethod
    async def fetch(
        cls,
        user_id: str,
        user_pw: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        view_state, event_validation = await cls._get_extra_payload()

        form = {
            "Txt_1": user_id,
            "Txt_2": user_pw,
            "__VIEWSTATE": view_state,
            "__EVENTVALIDATION": event_validation
        }
        return await HTTPClient.connector.post(
            cls.URL, headers=headers, body=form, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        """
        로그인 성공: status 302, location header 포함, 리다이렉트 메시지를 body에 포함
        로그인 실패: status 200, location header 미포함, alert 메시지를 body에 포함
        """
        # Login 성공
        if response.status == 302:
            iat = httpdate_to_unixtime(response.headers["date"])
            return ResourceData(
                data={"cookies": response.cookies, "iat": iat}, link=response.url
            )
        # Login 실패: 인트라넷 서버 과부하
        elif response.status == 503:
            return ErrorData(
                error={
                    "title": response.soup.find("h2").get_text(),
                    "error_message": response.soup.find("p").get_text()
                },
                link=response.url
            )
        # Login 실패: Common 한 오류
        else:
            alerts: List[str] = extract_alerts(response.soup)
            alert = alerts[0] if alerts else ""
            return ErrorData(
                error={"title": alert, "alert_messages": alerts}, link=response.url
            )


class StudentPhoto(IParser):
    URL: str = DOMAIN_NAME + "/SchoolRegMng/SR015.aspx"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        sid: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        query: Dict[str, str] = {"schNo": sid}
        query_string = urlencode(query)
        url = f"{cls.URL}?{query_string}"
        return await HTTPClient.connector.post(
            url, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        """
        사진을 불러온 경우:
            headers= {'transfer-encoding': 'chunked', 'content-type': 'image/jpeg', 'content-disposition': 'attachment;filename=image.jpeg'}
        사진을 불러오지 못한 경우:
            headers= {'transfer-encoding': 없음, 'content-type': 'text/html; charset=ks_c_5601-1987', 'content-disposition': 없음}
        """
        if response.headers["content-type"][:5] == "image":
            return ResourceData(data={"raw_image": response.raw}, link=response.url)
        else:
            return ErrorData(error={"title": "이미지를 불러올 수 없습니다."}, link=response.url)


class Chapel(ISemesterFetcher, IParser):
    URL: str = DOMAIN_NAME + "/StudentMng/SM050.aspx"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        semester: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await _post_with_semester(
            cls.URL, cookies, semester, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_summary(cls, response: Response) -> Dict[str, str]:
        soup = response.soup
        tbody = soup.find("tbody", attrs={"class": "viewbody"})
        if not tbody:
            raise ParsingError("채플 요약 테이블을 찾을 수 없습니다.", response)

        summary: Dict[str, str] = {}
        for th, td in zip(tbody.find_all("th"), tbody.find_all("td")):
            key = th.get_text(strip=True)
            value = td.get_text(strip=True)

            day_count = re.search(r"\d+", value)
            summary[key] = str(day_count.group()) if day_count else ""

        return summary

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List, List]:
        soup = response.soup
        thead = soup.find("thead", attrs={"class": "mhead"})
        tbody = soup.find("tbody", attrs={"class": "mbody"})

        return parse_table(response, thead, tbody)

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        summary = cls._parse_summary(response)
        head, body = cls._parse_main_table(response)

        return ResourceData(
            data={"summary": summary, "head": head, "body": body,},
            link=response.url,
            meta={
                "selected": response.etc["semester"].selected,
                "selectable": response.etc["semester"].selectable,
            },
        )


class Timetable(ISemesterFetcher, IParser):
    URL: str = DOMAIN_NAME + "/GradeMng/GD160.aspx"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        semester: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await _post_with_semester(
            cls.URL, cookies, semester, headers=headers, timeout=timeout, **kwargs
        )

    @staticmethod
    def _parse_contents(td: str, response: Response) -> Tuple:
        matching = re.match(
            r"(.+)?\(([^(]*)?\)(\d{2}:\d{2})\s*~\s*([0-9:]{,5})", td
        ) or re.match(r"(.+)?()(\d{2}:\d{2})\s*~\s*([0-9:]{,5})", td)
        if not matching:
            ParsingError("시간표 상세정보를 해석할 수 없습니다.", response)
        return matching.groups()

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List, List]:
        soup = response.soup
        thead = soup.find("thead", attrs={"class": "mhead"})
        tbody = soup.find("tbody", attrs={"class": "mbody"})
        result = [[], [], [], [], []]
        head, body = parse_table(response, thead, tbody)

        for row in body:
            for i, each in enumerate(row):
                if each:
                    result[i].append(cls._parse_contents(each, response))

        return head, result

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        head, body = cls._parse_main_table(response)
        return ResourceData(
            data={"head": head, "body": body},
            link=response.url,
            meta={
                "selected": response.etc["semester"].selected,
                "selectable": response.etc["semester"].selectable,
            },
        )


class Course(ISemesterFetcher, IParser):
    URL: str = DOMAIN_NAME + "/GradeMng/GD095.aspx"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        semester: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await _post_with_semester(
            cls.URL, cookies, semester, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List, List]:
        soup = response.soup
        thead = soup.find("thead", attrs={"class": "mhead"})
        tbody = soup.find("tbody", attrs={"class": "mbody"})

        return parse_table(response, thead, tbody)

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        head, body = cls._parse_main_table(response)
        return ResourceData(
            data={"head": head, "body": body},
            link=response.url,
            meta={
                "selected": response.etc["semester"].selected,
                "selectable": response.etc["semester"].selectable,
            },
        )


class GraduationExam(IParser):
    URL: str = DOMAIN_NAME + "/SchoolRegMng/SR050.aspx"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            cls.URL, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List, List]:
        soup = response.soup
        thead = soup.select_one('td > table > thead.mhead')
        tbody = soup.select_one('td > table > tbody.mbody')

        return parse_table(response, thead, tbody)

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        thead, body = cls._parse_main_table(response)
        return ResourceData(data={"head": thead, "body": body}, link=response.url)

      
class TotalAcceptanceStatus(IParser):
    URL: str = DOMAIN_NAME + "/GradeMng/GD010.aspx?viewRef=0"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            cls.URL, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )
    
    @classmethod
    def _parse_summary(cls, response: Response) -> Dict[str, str]:
        soup = response.soup
        table = soup.find("table", attrs={"class": "viewscore"})

        summary: Dict[str, str] = {}

        for td in table.find("tr").find_all('td'):
            if td.get_text() == "\xa0 ":
                continue
            key, value = td.get_text()[1:].split(" : ")
            summary[key] = value

        return summary

    @classmethod
    def _parse_head(cls, response: Response) -> List:
        soup = response.soup
        tbody = soup.find("tbody", attrs={"class": "viewbody"})
        head = []

        for i in range(3, 7):
            head.append(tbody.find('tr').find_all('th')[i].get_text())

        return head

    @classmethod
    def _parse_main_table(cls, response: Response) -> Dict[str, List]:
        soup = response.soup
        tbody = soup.find("tbody", attrs={"class": "viewbody"})
        key = ''
        courses_taken = defaultdict(list)

        for tr in tbody.find_all('tr'):
            division = []
            if tr.find('th', attrs={"rowspan": ""}):
                continue
            if tr.find('th'):
                key = tr.find('th').text
            for td in tr.find_all('td'):
                if td.get_text() == "":
                    continue
                division.append(td.get_text())
            courses_taken[key].append(division)

        return dict(courses_taken)
      
    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        summary = cls._parse_summary(response)
        head = cls._parse_head(response)
        body = cls._parse_main_table(response)
        return ResourceData(
            data={"summary": summary, "head": head, "body": body},
            link=response.url)


class Profile(IGeneralFetcher, IParser):
    URL: str = DOMAIN_NAME + "/SchoolRegMng/SR030.aspx"

    @staticmethod
    def validate_name(name: str) -> bool:
        return bool(re.search(r"^[가-힣]{2,}$", name))

    @staticmethod
    def validate_sid(univ_id: str) -> bool:
        return bool(re.search(r"^[a-zA-Z]{,1}\d{3,9}$", univ_id))

    @staticmethod
    def validate_major(major: str) -> bool:
        return bool(re.search(r"^[가-힣]{3,17}$", major))

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            cls.URL, headers=headers, cookies=cookies, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_sid(cls, response: Response) -> str:
        soup = response.soup
        sid_tag = soup.find("span", attrs={"id": "ctl00_ContentPlaceHolder1_Lab_3"})
        if not sid_tag:
            raise ParsingError("학번을 탐색할 수 없습니다.", response)

        sid: str = sid_tag.get_text(strip=True)
        if not cls.validate_sid(sid):
            raise ParsingError("올바르지 않은 학번입니다.", response)

        return sid

    @classmethod
    def _parse_name(cls, response: Response) -> str:
        soup = response.soup
        name_tag = soup.find("span", attrs={"id": "ctl00_ContentPlaceHolder1_Lab_4"})
        if not name_tag:
            raise ParsingError("성명을 탐색할 수 없습니다.", response)

        name = name_tag.get_text(strip=True)
        if not cls.validate_name(name):
            raise ParsingError("올바르지 않은 이름입니다.", response)

        return name

    @classmethod
    def _parse_major(cls, response: Response) -> str:
        soup = response.soup
        major_tag = soup.find("span", attrs={"id": "ctl00_ContentPlaceHolder1_Lab_5"})
        if not major_tag:
            raise ParsingError("전공을 탐색할 수 없습니다.", soup.original_markup)

        major = major_tag.get_text(strip=True)
        if not cls.validate_major(major):
            raise ParsingError("올바르지 않은 학과입니다.", response)

        return major

    @classmethod
    @_ParserPrecondition
    def parse_sid(cls, response: Response) -> APIResponseType:
        return ResourceData(data={"sid": cls._parse_sid(response)}, link=response.url)

    @classmethod
    @_ParserPrecondition
    def parse_name(cls, response: Response) -> APIResponseType:
        return ResourceData(data={"name": cls._parse_name(response)}, link=response.url)

    @classmethod
    @_ParserPrecondition
    def parse_major(cls, response: Response) -> APIResponseType:
        return ResourceData(
            data={"major": cls._parse_major(response)}, link=response.url
        )

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        sid = cls._parse_sid(response)
        name = cls._parse_name(response)
        major = cls._parse_major(response)
        return ResourceData(
            data={"sid": sid, "name": name, "major": major}, link=response.url
        )



""" 패키지에서 사용되는 예외

- RootError
    - RequestError
        - RequestTimeoutError
        - StatusError
            - ClientError
            - ServerError
    - ResponseError (paramter에 response 가 추가됨)
        - ParsingError
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .reqeust import Response

__all__ = (
    "RootError",
    "RequestError",
    "ResponseError",
    "RequestTimeoutError",
    "ClientError",
    "ServerError",
    "ParsingError",
)


class RootError(Exception):
    """ 모듈 최상위 에러 """


class RequestError(RootError):
    """ 요청과 관련된 top-level 에러 """


class ResponseError(RootError):
    """ 응답과 관련된 top-level 에러 """

    def __init__(self, message, response: "Response"):
        super().__init__(message)
        self.response = response


class RequestTimeoutError(RequestError):
    """ 요청 타임아웃 """


class StatusError(RequestError):
    """ HTTP 응답 코드가 400 이상인 경우 """


class ClientError(StatusError):
    """ 클라이언트 요청이 서버에서 처리할 수 없는 경우 (4xx) """


class ServerError(StatusError):
    """ 서버의 에러로 정상적인 반환이 불가한 경우 (5xx) """


class ParsingError(ResponseError):
    """ 구문 문석 에러 """



from typing import Optional, Dict, List
from abc import ABCMeta, abstractmethod
from dataclasses import asdict

from ..reqeust import Response
from ..exceptions import ParsingError
from .base import (
    HTTPClient,
    IParser,
    ILoginFetcher,
    APIResponseType,
    ErrorData,
    ResourceData,
    ParserPrecondition,
)
from .common import extract_alerts, httpdate_to_unixtime
from ._mileage import (
    translate_mileage_req,
    translate_statement_type,
    SearchParamData,
    StatementParamData,
)

__all__ = (
    "Login",
    "Search",
    "Statement",
)

DOMAIN_NAME: str = "https://asp.netusys.com"


class IParserPrecondition(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        """ 진행할 수 없는 사전조건인 경우 ErrorData, 그렇지 않은 경우 None """
        pass


_ParserPrecondition = ParserPrecondition(IParserPrecondition)


class _SessionExpiredChecker(IParserPrecondition):
    @staticmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        message_tag = response.soup.find("message")
        if message_tag and "세션정보" in message_tag.get_text(strip=True):
            return ErrorData(error={"title": "마일리지 세션이 만료되었습니다."}, link=response.url)

        return None


class Login(ILoginFetcher, IParser):
    URL: str = DOMAIN_NAME + "/login/login_check.jsp"
    DEFAULT_HEADERS: Dict[str, str] = {
        "referer": "https://asp.netusys.com/mobile/login/login_form.jsp?logoutFg=Y"
    }

    @classmethod
    async def fetch(
        cls,
        user_id: str,
        user_pw: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        headers = headers or {}
        headers.update(cls.DEFAULT_HEADERS)
        form: Dict[str, str] = {
            "user_id": user_id,
            "user_pwd": user_pw,
            "mac_addr": "",
            "AutoFg": "M",
            "appfg": "web",
            "logoutFg": "Y",
            "login_auto_serial": "",
        }

        response = await HTTPClient.connector.post(
            cls.URL, headers=headers, body=form, timeout=timeout, verify=False, **kwargs
        )
        return response

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        """
        성 공:          유효한 alert 없음, top.location.replace('/mobile/login/main.jsp?appfg=web&appYndHis=20200223201758');
        비밀번호 틀림:    유효한 alert 존재, top.location.replace('/mobile/login/login_form.jsp?logoutFg=Y');
        아이디가 틀림:    유효한 alert 존재, top.location.replace('/mobile/login/login_form.jsp?logoutFg=Y');
        """
        alerts = extract_alerts(response.soup)
        if alerts:
            return ErrorData(
                error={"title": alerts[0], "alert_messages": alerts}, link=response.url
            )
        iat = httpdate_to_unixtime(response.headers["date"])
        return ResourceData(
            data={"cookies": response.cookies, "iat": iat}, link=response.url
        )


def _parse_xml_data(response: Response) -> ResourceData:
    soup = response.soup

    data_container = soup.find("data")
    if not data_container:
        raise ParsingError("데이터를 찾을 수 없습니다.", response)

    total_rows_tag = soup.find("etc", attrs={"key": "total_rows"})
    if not total_rows_tag:
        raise ParsingError("전체 길이를 구할 수 없습니다.", response)

    total_row: str = total_rows_tag.get_text(strip=True)

    search_param = response.etc["req"]
    head: List[str] = translate_mileage_req(search_param.get_req())
    body: List[List[str]] = [
        [td.get_text(strip=True) for td in each.find_all("td")]
        for each in data_container.find_all("tr")
    ]
    page_num: str = search_param.get_page_num()

    if body:
        if len(head) != len(body[0]):
            raise ParsingError("데이터 헤드와 바디의 길이가 일치하지 않습니다.", response)

    return ResourceData(
        data={"head": head, "body": body},
        meta={"total_size": total_row, "current_size": len(body), "page_n": page_num,},
        link=response.url,
    )


class Search(IParser):
    URL: str = DOMAIN_NAME + "/ddd.sheetAction"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        search_param: Optional[SearchParamData] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        search_param = search_param or SearchParamData()
        response = await HTTPClient.connector.post(
            cls.URL,
            headers=headers,
            body=asdict(search_param),
            cookies=cookies,
            timeout=timeout,
            verify=False,
            **kwargs,
        )

        response.etc["req"]: SearchParamData = search_param

        return response

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        return _parse_xml_data(response)


class Statement(IParser):
    URL: str = DOMAIN_NAME + "/ddd.sheetAction"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        search_param: Optional[StatementParamData] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        search_param = search_param or StatementParamData()
        response = await HTTPClient.connector.post(
            cls.URL,
            headers=headers,
            body=asdict(search_param),
            cookies=cookies,
            timeout=timeout,
            verify=False,
            **kwargs,
        )

        response.etc["req"]: StatementParamData = search_param

        return response

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        result: ResourceData = _parse_xml_data(response)
        try:
            type_index = result.data["head"].index("구분")
        except ValueError:
            raise ParsingError("헤드에 마일리지변동 구분값이 존재하지 않습니다.", response)
        else:
            for each in result.data["body"]:
                each[type_index] = translate_statement_type(each[type_index])
        return result



""" HTTP Request/Response 추상화를 위한 클래스 """
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, Optional, Awaitable, Any, Callable, Type
from functools import wraps
import enum

from bs4 import BeautifulSoup

from ..exceptions import ClientError, ServerError

__all__ = (
    "Response",
    "HTTPRequestMethod",
    "BodyFormatter",
    "IRequestPostCondition",
    "BaseRequest",
    "DEFAULT_REQUEST_TIMEOUT",
)

DEFAULT_REQUEST_TIMEOUT: float = 30.0


@dataclass
class Response:
    """ HTTP Response 데이터 클래스 """

    status: int
    url: str
    reason: str = ""
    headers: Dict[str, str] = field(default_factory=dict)
    raw: bytes = b""
    text: str = ""
    cookies: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        self.headers = {key.lower(): value for key, value in self.headers.items()}

    def __bool__(self):
        return bool(self.status)

    @property
    def soup(self):
        try:
            return self._soup
        except AttributeError:
            self._soup = BeautifulSoup(self.text, features="html.parser")
            return self._soup

    @property
    def etc(self):
        try:
            return self._etc
        except AttributeError:
            self._etc: Dict[str, Any] = {}
            return self._etc


@enum.unique
class HTTPRequestMethod(enum.Enum):
    GET = "GET"
    POST = "POST"


@enum.unique
class BodyFormatter(enum.Enum):
    URL_ENCODE: str = "data"
    JSON: str = "json"


class IRequestPostCondition(metaclass=ABCMeta):
    """ HTTP request post-condition interface

    이 인터페이스를 상속한 모든 파생 클래스는 HTTP 응답 수신 이후 사후조건으로 실행됨
    """

    @staticmethod
    @abstractmethod
    def check(response: Response) -> None:
        pass


class _StatusCheck(IRequestPostCondition):
    """ Check HTTP response status code

    400번 이상의 상태 코드는 예외를 발생시킴
    4xx: ClientError
    5xx: ServerError
    """

    @staticmethod
    def check(response: Response) -> None:
        status = response.status
        n = status % 100
        if n == 4:
            raise ClientError(f"클라이언트 요청 오류입니다. -> 응답코드: {status}", response)
        elif n == 5:
            raise ServerError(f"서버 응답 오입니다. -> 응답코드: {status}", response)


class PostCondition:
    """ HTTP Request 요청의 사전/사후조건 처리를 위한 데코레이터

    사전조건: 타임아웃 설정
    사후조건: 응답 객체에 대한 사후조건 처리 (heckConditionBase 의 파생 클래스 실행)
    """

    def __init__(self, method: HTTPRequestMethod):
        self.method = method

    def __call__(
        self, request: Callable[..., Awaitable[Response]],
    ) -> Callable[..., Awaitable[Response]]:
        @wraps(request)
        async def check_condition(
            cls: Type["BaseRequest"], *args: Any, **kwargs: Any
        ) -> Response:
            response = await cls._request(self.method, *args, **kwargs)

            subclass: Type[IRequestPostCondition]
            for subclass in IRequestPostCondition.__subclasses__():
                subclass.check(response)
            return response

        return check_condition


class BaseRequest(metaclass=ABCMeta):
    """ HTTP Request abstract class

    파생 클래스는 _request 추상 메서드만 구현
    """

    @classmethod
    @PostCondition(HTTPRequestMethod.GET)
    async def get(
        cls,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        cookies: Dict[str, str] = None,
        verify: bool = True,
        allow_redirects: bool = False,
        timeout: Optional[float] = None,
        proxies: Optional[str] = None,
    ) -> Response:
        pass

    @classmethod
    @PostCondition(HTTPRequestMethod.POST)
    async def post(
        cls,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Dict[str, str]] = None,
        body_encoding: BodyFormatter = BodyFormatter.URL_ENCODE,
        cookies: Dict[str, str] = None,
        verify: bool = True,
        allow_redirects: bool = False,
        timeout: Optional[float] = None,
        proxies: Optional[str] = None,
    ) -> Response:
        pass

    @classmethod
    @abstractmethod
    async def _request(
        cls,
        method: HTTPRequestMethod,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Dict[str, str]] = None,
        body_encoding: BodyFormatter = BodyFormatter.URL_ENCODE,
        cookies: Dict[str, str] = None,
        verify: bool = True,
        allow_redirects: bool = False,
        timeout: Optional[float] = None,
        proxies: Optional[str] = None,
    ) -> Response:
        pass



from abc import ABCMeta, abstractmethod
from typing import Optional, Dict, List, Tuple
import re

from .base import (
    ILoginFetcher,
    IParser,
    APIResponseType,
    HTTPClient,
    ErrorData,
    ResourceData,
    SemesterData,
    IGeneralFetcher,
    ISemesterFetcher,
    ParserPrecondition,
)
from ..exceptions import ParsingError
from ..reqeust import Response
from .common import (
    httpdate_to_unixtime,
    extract_alerts,
    urlencode,
    SemesterConverter,
    parse_table,
)

__all__ = (
    "IParserPrecondition",
    "Login",
    "Profile",
    "CourseList",
    "Attendance",
)

DOMAIN_NAME: str = "https://lms.bible.ac.kr"  # with protocol


class IParserPrecondition(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        """ 진행할 수 없는 사전조건인 경우 ErrorData, 그렇지 않은 경우 None """
        pass


_ParserPrecondition = ParserPrecondition(IParserPrecondition)


class _SessionExpiredChecker(IParserPrecondition):
    @staticmethod
    def is_blocking(response: Response) -> Optional[ErrorData]:
        if "login" in response.headers.get("location", ""):
            return ErrorData(error={"title": "세션이 만료되었습니다."}, link=response.url,)

        return None


def _extract_semester(response: Response) -> SemesterData:
    soup = response.soup

    year_container = soup.find("select", attrs={"id": "year"})
    semester_container = soup.find("select", attrs={"id": "semester"})
    if not year_container:
        raise ParsingError("연도 셀렉트 태그를 불러올 수 없습니다.", response)
    elif not semester_container:
        raise ParsingError("학기 셀렉트 태그 불러올 수 없습니다.", response)

    try:
        year_selectables: List[str] = [
            opt.attrs["value"] for opt in year_container.find_all("option")
        ]
        year_selected: str = year_container.find("option", selected=True).attrs["value"]
        semester_selectables: List[str] = [
            opt.attrs["value"] for opt in semester_container.find_all_next("option")
        ]
        semester_selected: str = semester_container.find("option", selected=True).attrs[
            "value"
        ]

    except (KeyError, AttributeError) as e:
        raise ParsingError("학기 옵션 태그를 정상적으로 선택할 수 없습니다.", response) from e

    else:
        selected = SemesterConverter.lms_to_intranet(year_selected, semester_selected)
        selectable = [
            SemesterConverter.lms_to_intranet(year, semester)
            for year in year_selectables
            for semester in semester_selectables
        ]
    return SemesterData(selected=selected, selectable=selectable)


class Login(ILoginFetcher, IParser):
    URL: str = DOMAIN_NAME + "/login/index.php"
    LOGIN_ERROR = {
        "1": "현재, 브라우저의 쿠키가 작동하지 않습니다.",
        "2": "사용자 아이디: 이이디에는 영어소문자, 숫자, 밑줄( _ ), 하이폰( - ), 마침표( . ) 또는 @ 기호만을 쓸 수 있습니다.",
        "3": "아이디 또는 패스워드가 잘못 입력되었습니다.",
    }

    @classmethod
    async def fetch(
        cls,
        user_id: str,
        user_pw: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        form = {"username": user_id, "password": user_pw}
        return await HTTPClient.connector.post(
            cls.URL, headers=headers, body=form, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        # 로그인 실패
        if "location" not in response.headers:
            alerts = extract_alerts(response.soup)
            return ErrorData(
                error={"title": alerts[0], "alert_messages": alerts}, link=response.url,
            )
        m = re.search(r"errorcode=(\d+)", response.headers["location"])
        # 로그인 실패
        if m:
            error_code = m.group(1)
            return ErrorData(
                error={
                    "title": cls.LOGIN_ERROR.get(
                        error_code, f"login errorcode={error_code}"
                    ),
                    "code": error_code,
                },
                link=response.url,
            )
        # 로그인 성공
        else:
            iat = httpdate_to_unixtime(response.headers["date"])
            return ResourceData(
                data={"cookies": response.cookies, "iat": iat}, link=response.url
            )


class Profile(IGeneralFetcher, IParser):
    URL: str = DOMAIN_NAME + "/user/user_edit.php?lang=ko"

    @staticmethod
    def validate_name(name: str) -> bool:
        return bool(re.search(r"^[가-힣]{2,}$", name))

    @staticmethod
    def validate_sid(univ_id: str) -> bool:
        return bool(re.search(r"^[a-zA-Z]{,1}\d{3,9}$", univ_id))

    @staticmethod
    def validate_major(major: str) -> bool:
        return bool(re.search(r"^[가-힣]{3,17}$", major))

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            cls.URL, headers=headers, cookies=cookies, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_sid(cls, response: Response) -> str:
        soup = response.soup
        sid_container = soup.find("div", attrs={"id": "fitem_id_idnumber"})
        if not sid_container:
            raise ParsingError("학번 컨테이너를 찾을 수 없습니다.", response)

        sid_tag = sid_container.find("div", attrs={"class", "felement fstatic"})
        if not sid_tag:
            raise ParsingError("학번을 탐색할 수 없습니다.", response)

        sid: str = sid_tag.get_text(strip=True)
        if not cls.validate_sid(sid):
            raise ParsingError("올바르지 않은 학번입니다.", response)

        return sid

    @classmethod
    def _parse_name(cls, response: Response) -> str:
        soup = response.soup
        name_container = soup.find("div", attrs={"id": "fitem_id_firstname"})
        if not name_container:
            raise ParsingError("성명 컨테이너를 찾을 수 없습니다.", response)

        name_tag = name_container.find("input")
        if not name_tag:
            raise ParsingError("성명을 탐색할 수 없습니다.", response)

        name = name_tag.get("value", "").strip()
        if not cls.validate_name(name):
            raise ParsingError("올바르지 않은 이름입니다.", response)

        return name

    @classmethod
    def _parse_major(cls, response: Response) -> str:
        soup = response.soup
        major_container = soup.find("div", attrs={"id": "fitem_id_department"})
        if not major_container:
            raise ParsingError("전공 컨테이너를 찾을 수 없습니다.", soup.original_markup)

        major_tag = major_container.find("input")
        if not major_tag:
            raise ParsingError("전공을 탐색할 수 없습니다.", soup.original_markup)

        major = major_tag.get("value", "").strip()
        if not cls.validate_major(major):
            raise ParsingError("올바르지 않은 학과입니다.", response)

        return major

    @classmethod
    @_ParserPrecondition
    def parse_sid(cls, response: Response) -> APIResponseType:
        return ResourceData(data={"sid": cls._parse_sid(response)}, link=response.url)

    @classmethod
    @_ParserPrecondition
    def parse_name(cls, response: Response) -> APIResponseType:
        return ResourceData(data={"name": cls._parse_name(response)}, link=response.url)

    @classmethod
    @_ParserPrecondition
    def parse_major(cls, response: Response) -> APIResponseType:
        return ResourceData(
            data={"major": cls._parse_major(response)}, link=response.url
        )

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        sid = cls._parse_sid(response)
        name = cls._parse_name(response)
        major = cls._parse_major(response)
        return ResourceData(
            data={"sid": sid, "name": name, "major": major}, link=response.url
        )


class CourseList(ISemesterFetcher, IParser):
    URL: str = DOMAIN_NAME + "/local/ubion/user/index.php?lang=ko"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        semester: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        url = cls.URL

        if semester:
            query = SemesterConverter.intranet_to_lms(semester)
            query_string = urlencode(query)
            url = f"{url}&{query_string}"

        response = await HTTPClient.connector.get(
            url, headers=headers, cookies=cookies, timeout=timeout, **kwargs
        )
        response.etc["semester"] = semester
        return response

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        soup = response.soup

        courses: Dict[str, str] = {}
        for each in soup.find_all("a", attrs={"class": "coursefullname"}):
            course_code_matching = re.search(r"[?&]id=(\d+)", each["href"])
            if not course_code_matching:
                raise ParsingError("url에서 강좌 코드번호를 추출할 수 없습니다.", response)
            course_code = course_code_matching.group(1)
            lecture_name = re.sub(r"\[.*?\]", "", each.get_text(strip=True))
            courses[lecture_name] = course_code

        semester: SemesterData = _extract_semester(response)

        return ResourceData(
            data={"courses": courses},
            link=response.url,
            meta={"selected": semester.selected, "selectable": semester.selectable},
        )


class Attendance(IParser):
    URL: str = DOMAIN_NAME + "/local/ubattendance/my_status.php?lang=ko"

    @classmethod
    async def fetch(
        cls,
        cookies: Dict[str, str],
        course_code: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        query = {"id": course_code}
        query_string = urlencode(query)
        url = f"{cls.URL}&{query_string}"
        return await HTTPClient.connector.get(
            url, cookies=cookies, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def _parse_summary(cls, response: Response) -> Dict[str, str]:
        soup = response.soup
        summary_container = soup.find("div", attrs={"class": "course_info well"})
        if not summary_container:
            raise ParsingError("요약 정보를 찾을 수 없습니다.", response)

        result: Dict[str, str] = {
            each.contents[0]
            .get_text(strip=True): each.contents[1]
            .replace(":", "")
            .strip()
            for each in summary_container.find_all("li")
        }
        return result

    @classmethod
    def _parse_foot(cls, response: Response) -> Dict[str, str]:
        soup = response.soup
        tfoot = soup.find("tfoot")
        if not tfoot:
            raise ParsingError("테이블 foot이 존재하지 않습니다.", response)

        result = {}
        for each in tfoot.find_all("span"):
            key = each.contents[0].get_text(strip=True)
            value = re.search(r"\d+", each.contents[1])
            if not value:
                raise ParsingError("테이블 foot에서 수치를 발견할 수 없습니다.", response)
            result[key] = value.group()
        return result

    @classmethod
    def _parse_main_table(cls, response: Response) -> Tuple[List[str], List[List[str]]]:
        soup = response.soup

        table_container = soup.find(
            "table", attrs={"class": "attendance_my table table-bordered"}
        )
        if not table_container:
            raise ParsingError("메인 테이블을 찾을 수 없습니다.", response)

        thead = table_container.find("thead")
        tbody = table_container.find("tbody")
        return parse_table(response, thead, tbody)

    @classmethod
    def _precondidtion(cls, response: Response) -> Optional[ErrorData]:
        """
        응답 코드가 200인 경우 정상
        응답 코드가 303인 경우 수강 중이지 않은 강의
        응답 코드가 404인 경우 삭제된 강의
        """

        if response.status != 200:
            return ErrorData(
                error={"title": "출석 정보를 불러올 수 없는 강의입니다."}, link=response.url,
            )
        return None

    @classmethod
    @_ParserPrecondition
    def parse(cls, response: Response) -> APIResponseType:
        precondition = cls._precondidtion(response)
        if precondition is not None:
            return precondition
        else:
            summary = cls._parse_summary(response)
            head, body = cls._parse_main_table(response)
            foot = cls._parse_foot(response)
            return ResourceData(
                data={"summary": summary, "head": head, "body": body, "foot": foot},
                link=response.url,
            )



try:
    import aiohttp
except ImportError:
    raise ImportError("이 커넥터는 호출할 수 없습니다. 패키지를 설치해주세요.")

from typing import Optional, Dict
from http.cookies import SimpleCookie
from concurrent.futures._base import TimeoutError as _TimeoutError

from .base import (
    BaseRequest,
    Response,
    HTTPRequestMethod,
    BodyFormatter,
    DEFAULT_REQUEST_TIMEOUT,
)
from ..exceptions import RequestTimeoutError


class Request(BaseRequest):
    @classmethod
    async def _request(
        cls,
        method: HTTPRequestMethod,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Dict[str, str]] = None,
        body_encoding: BodyFormatter = BodyFormatter.URL_ENCODE,
        cookies: Dict[str, str] = None,
        verify: bool = True,
        allow_redirects: bool = False,
        timeout: Optional[float] = None,
        proxies: Optional[str] = None,
    ) -> Response:
        timeout = timeout or DEFAULT_REQUEST_TIMEOUT
        if proxies is not None:
            from aiohttp_socks import ProxyConnector

            proxies = ProxyConnector.from_url(proxies)
        async with aiohttp.ClientSession(connector=proxies) as session:
            try:
                async with session.request(
                    method.value,
                    url,
                    headers=headers,
                    cookies=cls._to_cookie_obj(cookies),
                    verify_ssl=verify,
                    allow_redirects=allow_redirects,
                    timeout=aiohttp.ClientTimeout(total=timeout),
                    **{body_encoding.value: body},
                ) as response:
                    cookies = cls._to_cookie_dict(response.cookies)
                    headers = dict(response.headers)
                    raw = await response.read()
                    try:
                        text = await response.text()
                    except UnicodeDecodeError:
                        text = ""

                    return Response(
                        response.status,
                        url,
                        response.reason,
                        headers,
                        raw,
                        text,
                        cookies,
                    )
            except _TimeoutError as e:
                raise RequestTimeoutError(f"요청시간이 경과하였습니다. -> {timeout}초") from e

    @staticmethod
    def _to_cookie_obj(cookies: Dict[str, str]) -> SimpleCookie:
        return SimpleCookie(cookies)

    @staticmethod
    def _to_cookie_dict(cookie: SimpleCookie) -> Dict[str, str]:
        return {morsel.key: morsel.value for morsel in cookie.values()}



__all__ = (
    "__title__",
    "__description__",
    "__url__",
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "__contributors__",
)

__title__ = "biblebot"
__description__ = "Scraper for KBU students"
__url__ = "https://github.com/rekyungmin/biblebot-scraper"
__version__ = "0.4.3"
__author__ = "Kyungmin Lee"
__author_email__ = "rekyungmin@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2023 Kyungmin Lee"
__contributors__ = "Yongki Kim <ygflove@likelion.org>",



import sys
from setuptools import setup, find_packages
from distutils.util import convert_path


if sys.version_info < (3, 7):
    raise RuntimeError("biblebot requires Python 3.7+")

with open("README.md", "r", encoding="UTF-8") as f:
    long_description = f.read()

about = {}
with open(convert_path("biblebot/__version__.py")) as f:
    exec(f.read(), about)


setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    maintainer=", ".join((f'{about["__author__"]} <{about["__author_email__"]}>', *about["__contributors__"])),
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about["__url__"],
    packages=find_packages(),
    install_requires=["beautifulsoup4 >= 4.8.0"],
    extras_require={"http": ["aiohttp[speedups]>=3.6.2"]},
    python_requires=">=3.7",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)



from dataclasses import dataclass
from typing import Optional, Dict, Union
import datetime
import unicodedata

from ..exceptions import ParsingError
from ..reqeust import Response
from .base import IParser, HTTPClient, APIResponseType, ResourceData
from .common import urlencode


__all__ = (
    "NoticeData",
    "NoticeArticle",
    "NoticeList",
    "MainNotice",
    "ScholarshipNotice",
    "IllipNotice",
)

DOMAIN_NAME: str = "https://www.bible.ac.kr"


@dataclass
class NoticeData:
    title: str
    author: str
    date: datetime.datetime
    content: str
    url: str


class NoticeArticle:
    @classmethod
    async def fetch(
        cls,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        return await HTTPClient.connector.get(
            url, headers=headers, timeout=timeout, **kwargs
        )

    @classmethod
    def parse(cls, response: Response) -> NoticeData:
        soup = response.soup

        header_container = soup.find("div", attrs={"class": "header"})
        if not header_container:
            raise ParsingError("공지사항 머리글 정보를 찾을 수 없습니다.", response)

        try:
            title = header_container.find("h5").get_text(strip=True)
            author = header_container.find("span", attrs={"rel": "author"}).get_text(
                strip=True
            )
            date = header_container.find("time").get_text(strip=True)
            content = soup.find("div", attrs={"class": "content"}).get_text(strip=True)

            converted_date = datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
            converted_content = unicodedata.normalize("NFKD", content).strip()
        except AttributeError as e:
            raise ParsingError("공지사항 컨텐츠 요소를 추출할 수 없습니다.", response) from e
        else:
            return NoticeData(
                title, author, converted_date, converted_content, response.url
            )


class NoticeList(IParser):
    """ 공지사항 리스트를 가져오는 클래스

    공지사항은 카테고리별로 여러 페이지로 나누어져 있지만, html 구성은 같기 때문에 본 클래스를 재사용할 수 있음
    상속받은 클래스에서 'URL' 클래스 변수 선언할 것
    """

    @classmethod
    async def fetch(
        cls,
        page: Union[int, str] = 1,
        search_keyword: Optional[str] = None,
        *,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Response:
        page = int(page)
        url = cls.URL + str(page)
        if search_keyword:
            query = {"keyword": search_keyword}
            query_string = urlencode(query)
            url = f"{url}?{query_string}"
        response = await HTTPClient.connector.get(
            url, headers=headers, timeout=timeout, **kwargs
        )
        response.etc["notice"] = {"page": page, "keyword": search_keyword}
        return response

    @classmethod
    def parse(cls, response: Response) -> APIResponseType:
        soup = response.soup
        table_container = soup.find(
            "ul", attrs={"data-role": "table", "class": "black"}
        )
        if not table_container:
            raise ParsingError("공지사항 컨테이너가 존재하지 않습니다.", response)

        rows = []
        for each in table_container.find_all("li", attrs={"class": "tbody"}):
            row = dict(
                seq=each.find("span", attrs={"class": "loopnum"}).get_text(strip=True),
                title=each.find("span", attrs={"class": "title"}).get_text(strip=True),
                author=each.find("span", attrs={"class": "name"}).get_text(strip=True),
                date=each.find("span", attrs={"class": "reg_date"}).get_text(
                    strip=True
                ),
                url=DOMAIN_NAME + each.find("a", href=True).attrs["href"],
            )
            rows.append(row)

        return ResourceData(
            data={"notice": rows}, meta=response.etc["notice"], link=response.url
        )


class MainNotice(NoticeList):
    URL: str = DOMAIN_NAME + "/ko/life/notice/list/"


class ScholarshipNotice(NoticeList):
    URL: str = DOMAIN_NAME + "/ko/life/tuition_notice/list/"


class IllipNotice(NoticeList):
    URL: str = DOMAIN_NAME + "/ko/illip/notice/list/"



try:
    import requests
except ImportError:
    raise ImportError("이 커넥터는 호출할 수 없습니다. 패키지를 설치해주세요.")

from typing import Optional, Dict
from functools import partial
import asyncio

import urllib3

from .base import (
    BaseRequest,
    Response,
    HTTPRequestMethod,
    BodyFormatter,
    DEFAULT_REQUEST_TIMEOUT,
)
from ..exceptions import RequestTimeoutError


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Request(BaseRequest):
    @classmethod
    async def _request(
        cls,
        method: HTTPRequestMethod,
        url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Dict[str, str]] = None,
        body_encoding: BodyFormatter = BodyFormatter.URL_ENCODE,
        cookies: Dict[str, str] = None,
        verify: bool = True,
        allow_redirects: bool = False,
        timeout: Optional[float] = None,
        proxies: Optional[str] = None,
    ) -> Response:
        timeout = timeout or DEFAULT_REQUEST_TIMEOUT
        proxies = None if proxies is None else {"http": proxies, "https": proxies}
        try:
            response: requests.models.Response = await asyncio.get_event_loop().run_in_executor(
                None,
                partial(
                    requests.request,
                    method.value,
                    url,
                    headers=headers,
                    cookies=cookies,
                    verify=verify,
                    allow_redirects=allow_redirects,
                    timeout=timeout,
                    proxies=proxies,
                    **{body_encoding.value: body},
                ),
            )
        except requests.exceptions.ConnectTimeout as e:
            raise RequestTimeoutError(f"요청시간이 경과하였습니다. -> {timeout}초") from e
        try:
            cookies = response.cookies.get_dict()
            return Response(
                response.status_code,
                url,
                response.reason,
                response.headers,
                response.content,
                response.text,
                cookies,
            )
        finally:
            response.close()
